from __future__ import annotations

import abc
import typing

from sqlalchemy import orm

from lime_uow import resources, exceptions

__all__ = (
    "SqlAlchemyUnitOfWork",
    "UnitOfWork",
)


R = typing.TypeVar("R", bound=resources.Resource[typing.Any])


class UnitOfWork(abc.ABC):
    def __init__(self):
        self.__resources: typing.Optional[
            typing.Dict[str, resources.Resource[typing.Any]]
        ] = None
        self.__resources_validated = False

    def __enter__(self) -> UnitOfWork:
        rs: typing.AbstractSet[resources.Resource[typing.Any]] = self.create_resources()
        if not self.__resources_validated:
            _check_for_duplicate_resource_names(rs)
            self.__resources_validated = True
        self.__resources = {r.__class__.resource_name(): r for r in rs}
        return self

    def __exit__(self, *args):
        self.rollback()
        self.__resources = None

    def get_resource(self, resource_type: typing.Type[R], /) -> R:
        if self.__resources is None:
            raise exceptions.OutsideTransactionError()
        else:
            return typing.cast(R, self.get_resource_by_name(resource_type.resource_name()))

    def get_resource_by_name(self, resource_name: str, /) -> resources.Resource[typing.Any]:
        if self.__resources is None:
            raise exceptions.OutsideTransactionError()
        else:
            try:
                return self.__resources[resource_name]
            except KeyError:
                raise exceptions.MissingResourceError(
                    resource_name=resource_name,
                    available_resources=self.__resources.keys(),
                )

    @abc.abstractmethod
    def create_resources(self) -> typing.AbstractSet[resources.Resource[typing.Any]]:
        raise NotImplementedError

    def rollback(self):
        err_messages: typing.List[str] = []
        if self.__resources is None:
            raise exceptions.OutsideTransactionError()
        else:
            for resource in self.__resources.values():
                try:
                    resource.rollback()
                except Exception as e:
                    err_messages.append(f"{resource.resource_name()}: {e}")

        if err_messages:
            err_msg_list = ", ".join(err_messages)
            err_msg = (
                f"The following errors occurred while performing a rollback on the"
                f" UnitOfWork: {err_msg_list}"
            )
            raise exceptions.RollbackError(err_msg)

    def save(self):
        # noinspection PyBroadException
        try:
            if self.__resources is None:
                raise exceptions.OutsideTransactionError()
            else:
                for resource in self.__resources.values():
                    resource.save()
        except:
            self.rollback()
            raise


def _check_for_duplicate_resource_names(rs: typing.Collection[R], /):
    names = [r.__class__.resource_name() for r in rs]
    duplicate_names = {name: ct for name in names if (ct := names.count(name)) > 1}
    if duplicate_names:
        raise exceptions.DuplicateResourceNames(duplicate_names)


class SqlAlchemyUnitOfWork(UnitOfWork, abc.ABC):
    def __init__(self, session_factory: orm.sessionmaker, /):
        super().__init__()
        self._session_factory = session_factory
        self._session: typing.Optional[orm.Session] = None

    @property
    def session(self) -> orm.Session:
        if self._session is None:
            raise exceptions.OutsideTransactionError()
        else:
            return self._session

    def __enter__(self) -> SqlAlchemyUnitOfWork:
        self._session = self._session_factory()
        return typing.cast(SqlAlchemyUnitOfWork, super().__enter__())

    def __exit__(self, *args):
        super().__exit__(*args)
        self.session.rollback()
        self.session.close()
        self._session = None



from __future__ import annotations

import abc
import typing

from sqlalchemy import orm

__all__ = (
    "DictRepository",
    "Repository",
    "Resource",
    "SqlAlchemyRepository",
)

from lime_uow import exceptions

T = typing.TypeVar("T", covariant=True)
E = typing.TypeVar("E")


class Resource(abc.ABC, typing.Generic[T]):
    @classmethod
    @abc.abstractmethod
    def resource_name(cls) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def rollback(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def save(self) -> None:
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        if other.__class__ is self.__class__:
            # noinspection PyTypeChecker
            return self.resource_name == typing.cast(Resource[typing.Any], other).resource_name
        else:
            return NotImplemented

    def __ne__(self, other: object) -> bool:
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        else:
            return not result

    def __hash__(self) -> int:
        return hash(self.resource_name)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.__class__.resource_name()}"


class Repository(Resource[E], abc.ABC, typing.Generic[E]):
    """Interface to access elements of a collection"""

    @abc.abstractmethod
    def add(self, item: E, /) -> E:
        raise NotImplementedError

    @abc.abstractmethod
    def all(self) -> typing.Iterable[E]:
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, item: E, /) -> E:
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, item: E, /) -> E:
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, item_id: typing.Any, /) -> E:
        raise NotImplementedError


class SqlAlchemyRepository(Repository[E], abc.ABC, typing.Generic[E]):
    def __init__(self, session: orm.Session):
        self._session = session

    def add(self, item: E, /) -> E:
        self.session.add(item)
        return item

    def all(self) -> typing.Generator[E, None, None]:
        return self.session.query(self.entity_type).all()

    def delete(self, item: E, /) -> E:
        self.session.delete(item)
        return item

    def get(self, item_id: typing.Any, /) -> E:
        return self.session.query(self.entity_type).get(item_id)

    @property
    @abc.abstractmethod
    def entity_type(self) -> typing.Type[E]:
        raise NotImplementedError

    @classmethod
    def resource_name(cls) -> str:
        descendant = _get_next_descendant_of(cls=cls, ancestor=SqlAlchemyRepository)
        if descendant is SqlAlchemyRepository:
            return cls.__name__
        else:
            return descendant.__name__

    def rollback(self) -> None:
        self.session.rollback()

    @property
    def session(self) -> orm.Session:
        return self._session

    def save(self) -> None:
        self.session.commit()

    def update(self, item: E, /) -> E:
        self.session.merge(item)
        return item

    def where(self, predicate: typing.Any, /) -> typing.List[E]:
        return self.session.query(self.entity_type).filter(predicate).all()


class DictRepository(Repository[E], abc.ABC, typing.Generic[E]):
    """Repository implementation based on a dictionary

    This exists primarily to make testing in client code simpler.  It was not designed for efficiency.
    """

    def __init__(
        self,
        *,
        initial_values: typing.Dict[typing.Hashable, E],
        key_fn: typing.Callable[[E], typing.Hashable],
    ):
        super().__init__()

        self._previous_state = initial_values
        self._current_state = initial_values.copy()
        self._key_fn = key_fn

        self.events: typing.List[typing.Tuple[str, typing.Dict[str, typing.Any]]] = []

    def rollback(self) -> None:
        self.events.append(("rollback", {}))
        self._current_state = self._previous_state.copy()

    def save(self) -> None:
        self.events.append(("save", {}))
        self._previous_state = self._current_state.copy()

    def add(self, item: E, /) -> E:
        self.events.append(("add", {"item": item}))
        key = self._key_fn(item)
        self._current_state[key] = item
        return item

    def all(self) -> typing.Iterable[E]:
        self.events.append(("all", {}))
        return list(self._current_state.values())

    def delete(self, item: E, /) -> E:
        self.events.append(("delete", {"item": item}))
        key = self._key_fn(item)
        del self._current_state[key]
        return item

    def update(self, item: E, /) -> E:
        self.events.append(("update", {"item": item}))
        key = self._key_fn(item)
        self._current_state[key] = item
        return item

    def get(self, item_id: typing.Any, /) -> E:
        self.events.append(("get", {"item_id": item_id}))
        return self._current_state[item_id]


def _get_next_descendant_of(
    cls: typing.Type[typing.Any], ancestor: typing.Type[typing.Any]
) -> typing.Type[typing.Any]:
    try:
        return next(c for c in cls.__mro__[1:] if ancestor in c.__mro__)
    except StopIteration:
        raise exceptions.NoCommonAncestor(
            f"Cannot find a common ancestor of {ancestor.__name__} for class {cls.__name__} in its "
            f"MRO.  The {cls.__name__}'s MRO is as follows: {', '.join(c.__name__ for c in cls.__mro__)}."
        )
